//
// Created by nicolas on 21/03/18.
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

void* (*_malloc)(size_t) = malloc;
void (*_free)(void*) = free;
#include "debug.h"

#define DBG_ALLOC_INC_SIZE  1024

struct st_alloc {
    const char* file;
    int line;
    size_t size;
    void *ptr;
};
size_t _allocations_count = 0;
size_t _allocations_size = 0;
struct st_alloc* _allocations = 0;

void* dbg_malloc(const char* file, int line, size_t size) {
    if (_allocations_count == _allocations_size) {
        _allocations = _allocations_size == 0 ? _malloc(DBG_ALLOC_INC_SIZE*sizeof(struct st_alloc)) : realloc(_allocations, (_allocations_size + DBG_ALLOC_INC_SIZE)*sizeof(struct st_alloc));
        memset(&_allocations[_allocations_size], 0, DBG_ALLOC_INC_SIZE * sizeof(struct st_alloc));
        _allocations_size += DBG_ALLOC_INC_SIZE;
    }
    void* ptr = _malloc(size);
    struct st_alloc* alloc = &_allocations[_allocations_count++];
    alloc->file = file;
    alloc->line = line;
    alloc->size = size;
    alloc->ptr = ptr;
    return ptr;
}

void dbg_free(void* ptr) {
    struct st_alloc* found = 0;
    for (size_t i = 0; i < _allocations_count; i++) {
        if (_allocations[i].ptr == ptr) found = &_allocations[i];
    }
    if (found) found->ptr = 0;

    _free(ptr);
}

void dbg_list_allocs() {
    size_t total = 0;
    size_t count = 0;
    printf("Memory leaks:\n");
    for (size_t i = 0; i < _allocations_count; i++) {
        if (_allocations[i].ptr == 0) continue;
        printf("%lu %p (%lu) @ %s:%d\n", i, _allocations[i].ptr, _allocations[i].size, _allocations[i].file, _allocations[i].line);
        total += _allocations[i].size;
        count++;
    }
    printf("Totals: nb=%lu memory=%lu\n", count, total);
}
